/*
 * Copyright (C) 2009-2010 Oracle Corporation
 *
 * All rights reserved
 */

package com.sun.network.rdp
{
    import flash.events.KeyboardEvent;
    import flash.ui.KeyLocation;
    import flash.utils.ByteArray;

    internal class KeyboardTranslator
    {
        private var mClient:RDPClient;
        
        /* Current keycode array. Assigned to a corresponding keyCodeXX array.
         * ActionScript reference: "Array assignment is by reference rather than by value."
         */
        private var mKeyCodeTable:Array;
        private var mCurrentLayout:String;

        /* The array maps a KeyboardEvent.keyCode to an index in a scancode sequences array.
         *
         * KeyboardEvent.keyCode is an index in the array. An element of the keyCode array is:
         * { numScancodes:uint, scancodesIndex:uint } for every "location".
         *
         * To retrieve:
         *    numScancodes = keyCodeUS[8 * keyCode + keyLocation * 2];
         *    indexScancodes = keyCodeUS[8 * keyCode + keyLocation * 2 + 1];
         *    switch (numScancodes)
         *    {
         *        case 1: scancodes[0] = scancodes1[indexScancodes]; return 1;
         *        case 2: scancodes[0] = scancodes2[2 * indexScancodes];
         *                scancodes[1] = scancodes2[2 * indexScancodes + 1]; return 2;
         *    }
         */
        private static const keyCodeUS:Array = 
        [
            /*                               numSc,index            */
            /* keycode           Std       Left     Right    NumPad */
            
            /* 0x00 (  0)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x01 (  1)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x02 (  2)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x03 (  3)  */    2,  4,    0, -1,   0, -1,   0, -1, /* Ctrl-Break (tested) */
            /* 0x04 (  4)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x05 (  5)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x06 (  6)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x07 (  7)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x08 (  8)  */    1, 13,    0, -1,   0, -1,   0, -1, /* BACKSPACE */
            /* 0x09 (  9)  */    1, 14,    0, -1,   0, -1,   0, -1, /* TAB */
            /* 0x0a ( 10)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x0b ( 11)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x0c ( 12)  */    1, 75,    0, -1,   0, -1,   1, 75, /* NumPad 5 (tested) */
            /* 0x0d ( 13)  */    1, 27,    0, -1,   0, -1,   2,  0, /* ENTER */
            /* 0x0e ( 14)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x0f ( 15)  */    1, -1,    0, -1,   0, -1,   0, -1, /* COMMAND */
            /* 0x10 ( 16)  */    1, 41,    1, 41,   1, 53,   0, -1, /* SHIFT */
            /* 0x11 ( 17)  */    1, 28,    1, 28,   2,  1,   0, -1, /* CONTROL */
            /* 0x12 ( 18)  */    1, 55,    1, 55,   2,  3,   0, -1, /* ALTERNATE */
            /* 0x13 ( 19)  */    0, -1,    0, -1,   0, -1,   0, -1, /* Break (tested) @todo which scancode? */
            /* 0x14 ( 20)  */    1, 57,    0, -1,   0, -1,   0, -1, /* CAPS_LOCK */
            /* 0x15 ( 21)  */    1, 68,    0, -1,   0, -1,   1, 68, /* NUMPAD */
            /* 0x16 ( 22)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x17 ( 23)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x18 ( 24)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x19 ( 25)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x1a ( 26)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x1b ( 27)  */    1,  0,    0, -1,   0, -1,   0, -1, /* ESCAPE */
            /* 0x1c ( 28)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x1d ( 29)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x1e ( 30)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x1f ( 31)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x20 ( 32)  */    1, 56,    0, -1,   0, -1,   0, -1, /* SPACE */
            /* 0x21 ( 33)  */    2,  7,    0, -1,   0, -1,   1, 72, /* PAGE_UP */
            /* 0x22 ( 34)  */    2, 12,    0, -1,   0, -1,   1, 80, /* PAGE_DOWN */
            /* 0x23 ( 35)  */    2, 10,    0, -1,   0, -1,   1, 78, /* END */
            /* 0x24 ( 36)  */    2,  5,    0, -1,   0, -1,   1, 70, /* HOME */
            /* 0x25 ( 37)  */    2,  8,    0, -1,   0, -1,   1, 74, /* LEFT */
            /* 0x26 ( 38)  */    2,  6,    0, -1,   0, -1,   1, 71, /* UP */
            /* 0x27 ( 39)  */    2,  9,    0, -1,   0, -1,   1, 76, /* RIGHT */
            /* 0x28 ( 40)  */    2, 11,    0, -1,   0, -1,   1, 79, /* DOWN */
            /* 0x29 ( 41)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x2a ( 42)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x2b ( 43)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x2c ( 44)  */    1, 83,    0, -1,   0, -1,   0, -1, /* Prt Scr/Sys Rq (tested) */
            /* 0x2d ( 45)  */    2, 13,    0, -1,   0, -1,   1, 81, /* INSERT */
            /* 0x2e ( 46)  */    2, 14,    0, -1,   0, -1,   1, 82, /* DELETE */
            /* 0x2f ( 47)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x30 ( 48)  */    1, 10,    0, -1,   0, -1,   0, -1, /* NUMBER_0 */
            /* 0x31 ( 49)  */    1,  1,    0, -1,   0, -1,   0, -1, /* NUMBER_1 */
            /* 0x32 ( 50)  */    1,  2,    0, -1,   0, -1,   0, -1, /* NUMBER_2 */
            /* 0x33 ( 51)  */    1,  3,    0, -1,   0, -1,   0, -1, /* NUMBER_3 */
            /* 0x34 ( 52)  */    1,  4,    0, -1,   0, -1,   0, -1, /* NUMBER_4 */
            /* 0x35 ( 53)  */    1,  5,    0, -1,   0, -1,   0, -1, /* NUMBER_5 */
            /* 0x36 ( 54)  */    1,  6,    0, -1,   0, -1,   0, -1, /* NUMBER_6 */
            /* 0x37 ( 55)  */    1,  7,    0, -1,   0, -1,   0, -1, /* NUMBER_7 */
            /* 0x38 ( 56)  */    1,  8,    0, -1,   0, -1,   0, -1, /* NUMBER_8 */
            /* 0x39 ( 57)  */    1,  9,    0, -1,   0, -1,   0, -1, /* NUMBER_9 */
            /* 0x3a ( 58)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x3b ( 59)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x3c ( 60)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x3d ( 61)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x3e ( 62)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x3f ( 63)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x40 ( 64)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x41 ( 65)  */    1, 29,    0, -1,   0, -1,   0, -1, /* A */
            /* 0x42 ( 66)  */    1, 47,    0, -1,   0, -1,   0, -1, /* B */
            /* 0x43 ( 67)  */    1, 45,    0, -1,   0, -1,   0, -1, /* C */
            /* 0x44 ( 68)  */    1, 31,    0, -1,   0, -1,   0, -1, /* D */
            /* 0x45 ( 69)  */    1, 17,    0, -1,   0, -1,   0, -1, /* E */
            /* 0x46 ( 70)  */    1, 32,    0, -1,   0, -1,   0, -1, /* F */
            /* 0x47 ( 71)  */    1, 33,    0, -1,   0, -1,   0, -1, /* G */
            /* 0x48 ( 72)  */    1, 34,    0, -1,   0, -1,   0, -1, /* H */
            /* 0x49 ( 73)  */    1, 22,    0, -1,   0, -1,   0, -1, /* I */
            /* 0x4a ( 74)  */    1, 35,    0, -1,   0, -1,   0, -1, /* J */
            /* 0x4b ( 75)  */    1, 36,    0, -1,   0, -1,   0, -1, /* K */
            /* 0x4c ( 76)  */    1, 37,    0, -1,   0, -1,   0, -1, /* L */
            /* 0x4d ( 77)  */    1, 49,    0, -1,   0, -1,   0, -1, /* M */
            /* 0x4e ( 78)  */    1, 48,    0, -1,   0, -1,   0, -1, /* N */
            /* 0x4f ( 79)  */    1, 23,    0, -1,   0, -1,   0, -1, /* O */
            /* 0x50 ( 80)  */    1, 24,    0, -1,   0, -1,   0, -1, /* P */
            /* 0x51 ( 81)  */    1, 15,    0, -1,   0, -1,   0, -1, /* Q */
            /* 0x52 ( 82)  */    1, 18,    0, -1,   0, -1,   0, -1, /* R */
            /* 0x53 ( 83)  */    1, 30,    0, -1,   0, -1,   0, -1, /* S */
            /* 0x54 ( 84)  */    1, 19,    0, -1,   0, -1,   0, -1, /* T */
            /* 0x55 ( 85)  */    1, 21,    0, -1,   0, -1,   0, -1, /* U */
            /* 0x56 ( 86)  */    1, 46,    0, -1,   0, -1,   0, -1, /* V */
            /* 0x57 ( 87)  */    1, 16,    0, -1,   0, -1,   0, -1, /* W */
            /* 0x58 ( 88)  */    1, 44,    0, -1,   0, -1,   0, -1, /* X */
            /* 0x59 ( 89)  */    1, 20,    0, -1,   0, -1,   0, -1, /* Y */
            /* 0x5a ( 90)  */    1, 43,    0, -1,   0, -1,   0, -1, /* Z */
            /* 0x5b ( 91)  */    2, 15,    0, -1,   0, -1,   0, -1, /* Left Windows key */
            /* 0x5c ( 92)  */    2, 16,    0, -1,   0, -1,   0, -1, /* Right Windows key */
            /* 0x5d ( 93)  */    2, 17,    0, -1,   0, -1,   0, -1, /* Menu */
            /* 0x5e ( 94)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x5f ( 95)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x60 ( 96)  */    1, 81,    0, -1,   0, -1,   1, 81, /* NUMPAD_0 */
            /* 0x61 ( 97)  */    1, 78,    0, -1,   0, -1,   1, 78, /* NUMPAD_1 */
            /* 0x62 ( 98)  */    1, 79,    0, -1,   0, -1,   1, 79, /* NUMPAD_2 */
            /* 0x63 ( 99)  */    1, 80,    0, -1,   0, -1,   1, 80, /* NUMPAD_3 */
            /* 0x64 (100)  */    1, 74,    0, -1,   0, -1,   1, 74, /* NUMPAD_4 */
            /* 0x65 (101)  */    1, 75,    0, -1,   0, -1,   1, 75, /* NUMPAD_5 */
            /* 0x66 (102)  */    1, 76,    0, -1,   0, -1,   1, 76, /* NUMPAD_6 */
            /* 0x67 (103)  */    1, 70,    0, -1,   0, -1,   1, 70, /* NUMPAD_7 */
            /* 0x68 (104)  */    1, 71,    0, -1,   0, -1,   1, 71, /* NUMPAD_8 */
            /* 0x69 (105)  */    1, 72,    0, -1,   0, -1,   1, 72, /* NUMPAD_9 */
            /* 0x6a (106)  */    1, 54,    0, -1,   0, -1,   1, 54, /* NUMPAD_MULTIPLY */
            /* 0x6b (107)  */    1, 77,    0, -1,   0, -1,   1, 77, /* NUMPAD_ADD */
            /* 0x6c (108)  */    2,  0,    0, -1,   0, -1,   2,  0, /* NUMPAD_ENTER */
            /* 0x6d (109)  */    1, 73,    0, -1,   0, -1,   1, 73, /* NUMPAD_SUBTRACT */
            /* 0x6e (110)  */    1, 82,    0, -1,   0, -1,   1, 82, /* NUMPAD_DECIMAL */
            /* 0x6f (111)  */    2,  2,    0, -1,   0, -1,   2,  2, /* NUMPAD_DIVIDE */
            /* 0x70 (112)  */    1, 58,    0, -1,   0, -1,   0, -1, /* F1 */
            /* 0x71 (113)  */    1, 59,    0, -1,   0, -1,   0, -1, /* F2 */
            /* 0x72 (114)  */    1, 60,    0, -1,   0, -1,   0, -1, /* F3 */
            /* 0x73 (115)  */    1, 61,    0, -1,   0, -1,   0, -1, /* F4 */
            /* 0x74 (116)  */    1, 62,    0, -1,   0, -1,   0, -1, /* F5 */
            /* 0x75 (117)  */    1, 63,    0, -1,   0, -1,   0, -1, /* F6 */
            /* 0x76 (118)  */    1, 64,    0, -1,   0, -1,   0, -1, /* F7 */
            /* 0x77 (119)  */    1, 65,    0, -1,   0, -1,   0, -1, /* F8 */
            /* 0x78 (120)  */    1, 66,    0, -1,   0, -1,   0, -1, /* F9 */
            /* 0x79 (121)  */    1, 67,    0, -1,   0, -1,   0, -1, /* F10 */
            /* 0x7a (122)  */    1, 84,    0, -1,   0, -1,   0, -1, /* F11 */
            /* 0x7b (123)  */    1, 85,    0, -1,   0, -1,   0, -1, /* F12 */
            /* 0x7c (124)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F13 */
            /* 0x7d (125)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F14 */
            /* 0x7e (126)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F15 */
            /* 0x7f (127)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x80 (128)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x81 (129)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x82 (130)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x83 (131)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x84 (132)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x85 (133)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x86 (134)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x87 (135)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x88 (136)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x89 (137)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8a (138)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8b (139)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8c (140)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8d (141)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8e (142)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x8f (143)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x90 (144)  */    1, 68,    0, -1,   0, -1,   1, 68, /* NumLock (tested) */
            /* 0x91 (145)  */    1, 69,    0, -1,   0, -1,   0, -1, /* ScrollLock (tested) */
            /* 0x92 (146)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x93 (147)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x94 (148)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x95 (149)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x96 (150)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x97 (151)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x98 (152)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x99 (153)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9a (154)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9b (155)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9c (156)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9d (157)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9e (158)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0x9f (159)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa0 (160)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa1 (161)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa2 (162)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa3 (163)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa4 (164)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa5 (165)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa6 (166)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa7 (167)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa8 (168)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xa9 (169)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xaa (170)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xab (171)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xac (172)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xad (173)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xae (174)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xaf (175)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb0 (176)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb1 (177)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb2 (178)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb3 (179)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb4 (180)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb5 (181)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb6 (182)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb7 (183)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb8 (184)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xb9 (185)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xba (186)  */    1, 38,    0, -1,   0, -1,   0, -1, /* SEMICOLON */
            /* 0xbb (187)  */    1, 12,    0, -1,   0, -1,   0, -1, /* EQUAL */
            /* 0xbc (188)  */    1, 50,    0, -1,   0, -1,   0, -1, /* COMMA */
            /* 0xbd (189)  */    1, 11,    0, -1,   0, -1,   1, 73, /* MINUS */
            /* 0xbe (190)  */    1, 51,    0, -1,   0, -1,   0, -1, /* PERIOD */
            /* 0xbf (191)  */    1, 52,    0, -1,   0, -1,   0, -1, /* SLASH */
            /* 0xc0 (192)  */    1, 40,    0, -1,   0, -1,   0, -1, /* BACKQUOTE */
            /* 0xc1 (193)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc2 (194)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc3 (195)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc4 (196)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc5 (197)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc6 (198)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc7 (199)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc8 (200)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xc9 (201)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xca (202)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xcb (203)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xcc (204)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xcd (205)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xce (206)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xcf (207)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd0 (208)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd1 (209)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd2 (210)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd3 (211)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd4 (212)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd5 (213)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd6 (214)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd7 (215)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd8 (216)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xd9 (217)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xda (218)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xdb (219)  */    1, 25,    0, -1,   0, -1,   0, -1, /* LEFTBRACKET [ */
            /* 0xdc (220)  */    1, 42,    0, -1,   0, -1,   0, -1, /* BACKSLASH */
            /* 0xdd (221)  */    1, 26,    0, -1,   0, -1,   0, -1, /* RIGHTBRACKET ] */
            /* 0xde (222)  */    1, 39,    0, -1,   0, -1,   0, -1, /* QUOTE */
            /* 0xdf (223)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe0 (224)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe1 (225)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe2 (226)  */    1, 86,    0, -1,   0, -1,   0, -1, /* BACKSLASH near left shift */
            /* 0xe3 (227)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe4 (228)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe5 (229)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe6 (230)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe7 (231)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe8 (232)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xe9 (233)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xea (234)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xeb (235)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xec (236)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xed (237)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xee (238)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xef (239)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf0 (240)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf1 (241)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf2 (242)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf3 (243)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf4 (244)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf5 (245)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf6 (246)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf7 (247)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf8 (248)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xf9 (249)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xfa (250)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xfb (251)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xfc (252)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xfd (253)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xfe (254)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* 0xff (255)  */    0, -1,    0, -1,   0, -1,   0, -1,
        ];

        private static const keyCodeDE:Array = 
        [
            /*                               numSc,index            */
            /* keycode           Std       Left     Right    NumPad */
            
            /* de 0x00 (  0)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x01 (  1)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x02 (  2)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x03 (  3)  */    2,  4,    0, -1,   0, -1,   0, -1, /* Ctrl-Break (tested) */
            /* de 0x04 (  4)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x05 (  5)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x06 (  6)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x07 (  7)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x08 (  8)  */    1, 13,    0, -1,   0, -1,   0, -1, /* BACKSPACE */
            /* de 0x09 (  9)  */    1, 14,    0, -1,   0, -1,   0, -1, /* TAB */
            /* de 0x0a ( 10)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x0b ( 11)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x0c ( 12)  */    1, 75,    0, -1,   0, -1,   1, 75, /* NumPad 5 (tested) */
            /* de 0x0d ( 13)  */    1, 27,    0, -1,   0, -1,   2,  0, /* ENTER */
            /* de 0x0e ( 14)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x0f ( 15)  */    1, -1,    0, -1,   0, -1,   0, -1, /* COMMAND */
            /* de 0x10 ( 16)  */    1, 41,    1, 41,   1, 53,   0, -1, /* SHIFT */
            /* de 0x11 ( 17)  */    1, 28,    1, 28,   2,  1,   0, -1, /* CONTROL */
            /* de 0x12 ( 18)  */    1, 55,    1, 55,   2,  3,   0, -1, /* ALTERNATE */
            /* de 0x13 ( 19)  */    0, -1,    0, -1,   0, -1,   0, -1, /* Break (tested) @todo which scancode? */
            /* de 0x14 ( 20)  */    1, 57,    0, -1,   0, -1,   0, -1, /* CAPS_LOCK */
            /* de 0x15 ( 21)  */    1, 68,    0, -1,   0, -1,   1, 68, /* NUMPAD */
            /* de 0x16 ( 22)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x17 ( 23)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x18 ( 24)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x19 ( 25)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x1a ( 26)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x1b ( 27)  */    1,  0,    0, -1,   0, -1,   0, -1, /* ESCAPE */
            /* de 0x1c ( 28)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x1d ( 29)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x1e ( 30)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x1f ( 31)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x20 ( 32)  */    1, 56,    0, -1,   0, -1,   0, -1, /* SPACE */
            /* de 0x21 ( 33)  */    2,  7,    0, -1,   0, -1,   1, 72, /* PAGE_UP */
            /* de 0x22 ( 34)  */    2, 12,    0, -1,   0, -1,   1, 80, /* PAGE_DOWN */
            /* de 0x23 ( 35)  */    2, 10,    0, -1,   0, -1,   1, 78, /* END */
            /* de 0x24 ( 36)  */    2,  5,    0, -1,   0, -1,   1, 70, /* HOME */
            /* de 0x25 ( 37)  */    2,  8,    0, -1,   0, -1,   1, 74, /* LEFT */
            /* de 0x26 ( 38)  */    2,  6,    0, -1,   0, -1,   1, 71, /* UP */
            /* de 0x27 ( 39)  */    2,  9,    0, -1,   0, -1,   1, 76, /* RIGHT */
            /* de 0x28 ( 40)  */    2, 11,    0, -1,   0, -1,   1, 79, /* DOWN */
            /* de 0x29 ( 41)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x2a ( 42)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x2b ( 43)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x2c ( 44)  */    1, 83,    0, -1,   0, -1,   0, -1, /* Prt Scr/Sys Rq (tested) */
            /* de 0x2d ( 45)  */    2, 13,    0, -1,   0, -1,   1, 81, /* INSERT */
            /* de 0x2e ( 46)  */    2, 14,    0, -1,   0, -1,   1, 82, /* DELETE */
            /* de 0x2f ( 47)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x30 ( 48)  */    1, 10,    0, -1,   0, -1,   0, -1, /* NUMBER_0 */
            /* de 0x31 ( 49)  */    1,  1,    0, -1,   0, -1,   0, -1, /* NUMBER_1 */
            /* de 0x32 ( 50)  */    1,  2,    0, -1,   0, -1,   0, -1, /* NUMBER_2 */
            /* de 0x33 ( 51)  */    1,  3,    0, -1,   0, -1,   0, -1, /* NUMBER_3 */
            /* de 0x34 ( 52)  */    1,  4,    0, -1,   0, -1,   0, -1, /* NUMBER_4 */
            /* de 0x35 ( 53)  */    1,  5,    0, -1,   0, -1,   0, -1, /* NUMBER_5 */
            /* de 0x36 ( 54)  */    1,  6,    0, -1,   0, -1,   0, -1, /* NUMBER_6 */
            /* de 0x37 ( 55)  */    1,  7,    0, -1,   0, -1,   0, -1, /* NUMBER_7 */
            /* de 0x38 ( 56)  */    1,  8,    0, -1,   0, -1,   0, -1, /* NUMBER_8 */
            /* de 0x39 ( 57)  */    1,  9,    0, -1,   0, -1,   0, -1, /* NUMBER_9 */
            /* de 0x3a ( 58)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x3b ( 59)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x3c ( 60)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x3d ( 61)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x3e ( 62)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x3f ( 63)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x40 ( 64)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x41 ( 65)  */    1, 29,    0, -1,   0, -1,   0, -1, /* A */
            /* de 0x42 ( 66)  */    1, 47,    0, -1,   0, -1,   0, -1, /* B */
            /* de 0x43 ( 67)  */    1, 45,    0, -1,   0, -1,   0, -1, /* C */
            /* de 0x44 ( 68)  */    1, 31,    0, -1,   0, -1,   0, -1, /* D */
            /* de 0x45 ( 69)  */    1, 17,    0, -1,   0, -1,   0, -1, /* E */
            /* de 0x46 ( 70)  */    1, 32,    0, -1,   0, -1,   0, -1, /* F */
            /* de 0x47 ( 71)  */    1, 33,    0, -1,   0, -1,   0, -1, /* G */
            /* de 0x48 ( 72)  */    1, 34,    0, -1,   0, -1,   0, -1, /* H */
            /* de 0x49 ( 73)  */    1, 22,    0, -1,   0, -1,   0, -1, /* I */
            /* de 0x4a ( 74)  */    1, 35,    0, -1,   0, -1,   0, -1, /* J */
            /* de 0x4b ( 75)  */    1, 36,    0, -1,   0, -1,   0, -1, /* K */
            /* de 0x4c ( 76)  */    1, 37,    0, -1,   0, -1,   0, -1, /* L */
            /* de 0x4d ( 77)  */    1, 49,    0, -1,   0, -1,   0, -1, /* M */
            /* de 0x4e ( 78)  */    1, 48,    0, -1,   0, -1,   0, -1, /* N */
            /* de 0x4f ( 79)  */    1, 23,    0, -1,   0, -1,   0, -1, /* O */
            /* de 0x50 ( 80)  */    1, 24,    0, -1,   0, -1,   0, -1, /* P */
            /* de 0x51 ( 81)  */    1, 15,    0, -1,   0, -1,   0, -1, /* Q */
            /* de 0x52 ( 82)  */    1, 18,    0, -1,   0, -1,   0, -1, /* R */
            /* de 0x53 ( 83)  */    1, 30,    0, -1,   0, -1,   0, -1, /* S */
            /* de 0x54 ( 84)  */    1, 19,    0, -1,   0, -1,   0, -1, /* T */
            /* de 0x55 ( 85)  */    1, 21,    0, -1,   0, -1,   0, -1, /* U */
            /* de 0x56 ( 86)  */    1, 46,    0, -1,   0, -1,   0, -1, /* V */
            /* de 0x57 ( 87)  */    1, 16,    0, -1,   0, -1,   0, -1, /* W */
            /* de 0x58 ( 88)  */    1, 44,    0, -1,   0, -1,   0, -1, /* X */
            /* de 0x59 ( 89)  */    1, 43,    0, -1,   0, -1,   0, -1, /* Y */
            /* de 0x5a ( 90)  */    1, 20,    0, -1,   0, -1,   0, -1, /* Z */
            /* de 0x5b ( 91)  */    2, 15,    0, -1,   0, -1,   0, -1, /* Left Windows key */
            /* de 0x5c ( 92)  */    2, 16,    0, -1,   0, -1,   0, -1, /* Right Windows key */
            /* de 0x5d ( 93)  */    2, 17,    0, -1,   0, -1,   0, -1, /* Menu */
            /* de 0x5e ( 94)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x5f ( 95)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x60 ( 96)  */    1, 81,    0, -1,   0, -1,   1, 81, /* NUMPAD_0 */
            /* de 0x61 ( 97)  */    1, 78,    0, -1,   0, -1,   1, 78, /* NUMPAD_1 */
            /* de 0x62 ( 98)  */    1, 79,    0, -1,   0, -1,   1, 79, /* NUMPAD_2 */
            /* de 0x63 ( 99)  */    1, 80,    0, -1,   0, -1,   1, 80, /* NUMPAD_3 */
            /* de 0x64 (100)  */    1, 74,    0, -1,   0, -1,   1, 74, /* NUMPAD_4 */
            /* de 0x65 (101)  */    1, 75,    0, -1,   0, -1,   1, 75, /* NUMPAD_5 */
            /* de 0x66 (102)  */    1, 76,    0, -1,   0, -1,   1, 76, /* NUMPAD_6 */
            /* de 0x67 (103)  */    1, 70,    0, -1,   0, -1,   1, 70, /* NUMPAD_7 */
            /* de 0x68 (104)  */    1, 71,    0, -1,   0, -1,   1, 71, /* NUMPAD_8 */
            /* de 0x69 (105)  */    1, 72,    0, -1,   0, -1,   1, 72, /* NUMPAD_9 */
            /* de 0x6a (106)  */    1, 54,    0, -1,   0, -1,   1, 54, /* NUMPAD_MULTIPLY */
            /* de 0x6b (107)  */    1, 77,    0, -1,   0, -1,   1, 77, /* NUMPAD_ADD */
            /* de 0x6c (108)  */    2,  0,    0, -1,   0, -1,   2,  0, /* NUMPAD_ENTER */
            /* de 0x6d (109)  */    1, 73,    0, -1,   0, -1,   1, 73, /* NUMPAD_SUBTRACT */
            /* de 0x6e (110)  */    1, 82,    0, -1,   0, -1,   1, 82, /* NUMPAD_DECIMAL */
            /* de 0x6f (111)  */    2,  2,    0, -1,   0, -1,   2,  2, /* NUMPAD_DIVIDE */
            /* de 0x70 (112)  */    1, 58,    0, -1,   0, -1,   0, -1, /* F1 */
            /* de 0x71 (113)  */    1, 59,    0, -1,   0, -1,   0, -1, /* F2 */
            /* de 0x72 (114)  */    1, 60,    0, -1,   0, -1,   0, -1, /* F3 */
            /* de 0x73 (115)  */    1, 61,    0, -1,   0, -1,   0, -1, /* F4 */
            /* de 0x74 (116)  */    1, 62,    0, -1,   0, -1,   0, -1, /* F5 */
            /* de 0x75 (117)  */    1, 63,    0, -1,   0, -1,   0, -1, /* F6 */
            /* de 0x76 (118)  */    1, 64,    0, -1,   0, -1,   0, -1, /* F7 */
            /* de 0x77 (119)  */    1, 65,    0, -1,   0, -1,   0, -1, /* F8 */
            /* de 0x78 (120)  */    1, 66,    0, -1,   0, -1,   0, -1, /* F9 */
            /* de 0x79 (121)  */    1, 67,    0, -1,   0, -1,   0, -1, /* F10 */
            /* de 0x7a (122)  */    1, 84,    0, -1,   0, -1,   0, -1, /* F11 */
            /* de 0x7b (123)  */    1, 85,    0, -1,   0, -1,   0, -1, /* F12 */
            /* de 0x7c (124)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F13 */
            /* de 0x7d (125)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F14 */
            /* de 0x7e (126)  */    1, -1,    0, -1,   0, -1,   0, -1, /* F15 */
            /* de 0x7f (127)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x80 (128)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x81 (129)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x82 (130)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x83 (131)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x84 (132)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x85 (133)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x86 (134)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x87 (135)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x88 (136)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x89 (137)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8a (138)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8b (139)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8c (140)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8d (141)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8e (142)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x8f (143)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x90 (144)  */    1, 68,    0, -1,   0, -1,   1, 68, /* NumLock (tested) */
            /* de 0x91 (145)  */    1, 69,    0, -1,   0, -1,   0, -1, /* ScrollLock (tested) */
            /* de 0x92 (146)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x93 (147)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x94 (148)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x95 (149)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x96 (150)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x97 (151)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x98 (152)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x99 (153)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9a (154)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9b (155)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9c (156)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9d (157)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9e (158)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0x9f (159)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa0 (160)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa1 (161)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa2 (162)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa3 (163)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa4 (164)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa5 (165)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa6 (166)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa7 (167)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa8 (168)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xa9 (169)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xaa (170)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xab (171)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xac (172)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xad (173)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xae (174)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xaf (175)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb0 (176)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb1 (177)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb2 (178)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb3 (179)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb4 (180)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb5 (181)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb6 (182)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb7 (183)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb8 (184)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xb9 (185)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xba (186)  */    1, 25,    0, -1,   0, -1,   0, -1, /* [{ */
            /* de 0xbb (187)  */    1, 26,    0, -1,   0, -1,   0, -1, /* ]} */
            /* de 0xbc (188)  */    1, 50,    0, -1,   0, -1,   0, -1, /* COMMA */
            /* de 0xbd (189)  */    1, 52,    0, -1,   0, -1,   1, 73, /* /? */
            /* de 0xbe (190)  */    1, 51,    0, -1,   0, -1,   0, -1, /* PERIOD */
            /* de 0xbf (191)  */    1, 42,    0, -1,   0, -1,   0, -1, /* \| on the right near Enter */
            /* de 0xc0 (192)  */    1, 38,    0, -1,   0, -1,   0, -1, /* ;: */
            /* de 0xc1 (193)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc2 (194)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc3 (195)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc4 (196)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc5 (197)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc6 (198)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc7 (199)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc8 (200)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xc9 (201)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xca (202)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xcb (203)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xcc (204)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xcd (205)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xce (206)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xcf (207)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd0 (208)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd1 (209)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd2 (210)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd3 (211)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd4 (212)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd5 (213)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd6 (214)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd7 (215)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd8 (216)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xd9 (217)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xda (218)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xdb (219)  */    1, 11,    0, -1,   0, -1,   0, -1, /* -_ */
            /* de 0xdc (220)  */    1, 40,    0, -1,   0, -1,   0, -1, /* `~ on the left of 1 */
            /* de 0xdd (221)  */    1, 12,    0, -1,   0, -1,   0, -1, /* =+ */
            /* de 0xde (222)  */    1, 39,    0, -1,   0, -1,   0, -1, /* QUOTE */
            /* de 0xdf (223)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe0 (224)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe1 (225)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe2 (226)  */    1, 86,    0, -1,   0, -1,   0, -1, /* BACKSLASH near left shift */
            /* de 0xe3 (227)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe4 (228)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe5 (229)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe6 (230)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe7 (231)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe8 (232)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xe9 (233)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xea (234)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xeb (235)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xec (236)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xed (237)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xee (238)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xef (239)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf0 (240)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf1 (241)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf2 (242)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf3 (243)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf4 (244)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf5 (245)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf6 (246)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf7 (247)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf8 (248)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xf9 (249)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xfa (250)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xfb (251)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xfc (252)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xfd (253)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xfe (254)  */    0, -1,    0, -1,   0, -1,   0, -1,
            /* de 0xff (255)  */    0, -1,    0, -1,   0, -1,   0, -1,
        ];

        /*
         * Arrays of scancodes referenced in the keyCodes table by index.
         */
        private static const scancodes2:Array = 
        [
        /*    sc           US key,     index */
            0xe0, 0x1c, /* (Enter)       0
                         * Note: NumPad Enter produces keyCode 13 and keyLocation 0 just like normal Enter
                         * in Flash here. Make the array element anyway since Flash defines a NUMPAD_ENTER keycode.
                         */
            0xe0, 0x1d, /* (RCtrl)       1 */
            0xe0, 0x35, /* (Keypad-/)    2 */
            0xe0, 0x38, /* (RAlt)        3 @todo intercepted by browser/player. */
            0xe0, 0x46, /* (Ctrl-Break)  4 */
            0xe0, 0x47, /* (Grey Home)   5 */
            0xe0, 0x48, /* (Grey Up)     6 */
            0xe0, 0x49, /* (Grey PgUp)   7 */
            0xe0, 0x4b, /* (Grey Left)   8 */
            0xe0, 0x4d, /* (Grey Right)  9 */
            0xe0, 0x4f, /* (Grey End)   10 */
            0xe0, 0x50, /* (Grey Down)  11 */
            0xe0, 0x51, /* (Grey PgDn)  12 */
            0xe0, 0x52, /* (Grey Insert) 13 */
            0xe0, 0x53, /* (Grey Delete) 14 */
            0xe0, 0x5b, /* (LWin)       15 */
            0xe0, 0x5c, /* (RWin)       16 */
            0xe0, 0x5d, /* (Menu)       17 */
            
            
        ];
        
        private static const scancodes1:Array = 
        [
        /*    sc,    US key,index*/
            0x01, /* (Esc)  0 */
            0x02, /* (1!)   1 */
            0x03, /* (2@)   2 */
            0x04, /* (3#)   3 */
            0x05, /* (4$)   4 */
            0x06, /* (5%)   5 */
            0x07, /* (6^)   6 */
            0x08, /* (7&)   7 */
            0x09, /* (8*)   8 */
            0x0a, /* (9()   9 */
            0x0b, /* (0))  10 */
            0x0c, /* (-_)  11 */
            0x0d, /* (=+)  12 */
            0x0e, /* (Backspace) 13 */
            0x0f, /* (Tab) 14 */
            0x10, /* (Q)   15 */
            0x11, /* (W)   16 */
            0x12, /* (E)   17 */
            0x13, /* (R)   18 */
            0x14, /* (T)   19 */
            0x15, /* (Y)   20 */
            0x16, /* (U)   21 */
            0x17, /* (I)   22 */
            0x18, /* (O)   23 */
            0x19, /* (P)   24 */
            0x1a, /* ([{)  25 */
            0x1b, /* (]})  26 */
            0x1c, /* (Enter) 27 */
            0x1d, /* (LCtrl) 28 */
            0x1e, /* (A)   29 */
            0x1f, /* (S)   30 */
            0x20, /* (D)   31 */
            0x21, /* (F)   32 */
            0x22, /* (G)   33 */
            0x23, /* (H)   34 */
            0x24, /* (J)   35 */
            0x25, /* (K)   36 */
            0x26, /* (L)   37 */
            0x27, /* (;:)  38 */
            0x28, /* ('")  39 */
            0x29, /* (`~)  40 */
            0x2a, /* (LShift) 41 */
            0x2b, /* (\|)  42 */
            0x2c, /* (Z)   43 */
            0x2d, /* (X)   44 */
            0x2e, /* (C)   45 */
            0x2f, /* (V)   46 */
            0x30, /* (B)   47 */
            0x31, /* (N)   48 */
            0x32, /* (M)   49 */
            0x33, /* (,<)  50 */
            0x34, /* (.>)  51 */
            0x35, /* (/?)  52 */
            0x36, /* (RShift) 53 */
            0x37, /* (Keypad-*) 54 */
            0x38, /* (LAlt) 55 @todo Alt is intercepted by the browser/player */
            0x39, /* (Space bar) 56 */
            0x3a, /* (CapsLock) 57 */
            0x3b, /* (F1)  58 */
            0x3c, /* (F2)  59 */
            0x3d, /* (F3)  60 */
            0x3e, /* (F4)  61 */
            0x3f, /* (F5)  62 */
            0x40, /* (F6)  63 */
            0x41, /* (F7)  64 */
            0x42, /* (F8)  65 */
            0x43, /* (F9)  66 */
            0x44, /* (F10) 67 */
            0x45, /* (NumLock) 68 */
            0x46, /* (ScrollLock) 69 */
            0x47, /* (Keypad-7/Home) 70 */
            0x48, /* (Keypad-8/Up)   71 */
            0x49, /* (Keypad-9/PgUp) 72 */
            0x4a, /* (Keypad--)      73 */
            0x4b, /* (Keypad-4/Left) 74 */
            0x4c, /* (Keypad-5)      75 */
            0x4d, /* (Keypad-6/Right) 76 */
            0x4e, /* (Keypad-+)      77 */
            0x4f, /* (Keypad-1/End)  78 */
            0x50, /* (Keypad-2/Down) 79 */
            0x51, /* (Keypad-3/PgDn) 80 */
            0x52, /* (Keypad-0/Ins)  81 */
            0x53, /* (Keypad-./Del)  82 */
            0x54, /* (Alt-SysRq)     83 @todo sometimes not passed to flash */
            0x57, /* (F11) 84 */
            0x58, /* (F12) 85 */
            0x56, /* (\|)  86 near Left Shift */
        ];
        
        private var mfRightShiftPressed:Boolean;

        public function KeyboardTranslator(aClient:RDPClient)
        {
            mClient = aClient;

            mClient.debug().msg("KeyboardTranslator\n");
            
            mfRightShiftPressed = false;
            
            /* By default use US keyboard. */
            SetLayout("en");
        }
        
        private function scancode (scancodes:ByteArray, event:KeyboardEvent, keyCodeArray:Array):uint
        {
            /* US event -> US scancodes. */
            if (event.keyCode >= 256)
            {
                return 0; /* means: no scancodes, ignore the event. */
            }
            
            var keyLocation:uint;
            
            switch (event.keyLocation)
            {
                default:
                case KeyLocation.STANDARD:
                    keyLocation = 0;
                    break;
                    
                case KeyLocation.LEFT:
                    keyLocation = 1;
                    break;
                    
                case KeyLocation.RIGHT:
                    keyLocation = 2;
                    break;
                    
                case KeyLocation.NUM_PAD:
                    keyLocation = 3;
                    break;
            }
            
            var indexKeyRecord:uint = 8 * event.keyCode + 2 * keyLocation;
            
            var numScancodes:uint = keyCodeArray[indexKeyRecord];
            var indexScancodes:uint = keyCodeArray[indexKeyRecord + 1];
            
            if (indexScancodes == -1)
            {
                return 0;
            }
            
            if (numScancodes == 1)
            {
                if (indexScancodes >= scancodes1.length)
                {
                    return 0;
                }
                
                scancodes[0] = scancodes1[indexScancodes];
            }
            else if (numScancodes == 2)
            {
                if (indexScancodes >= scancodes2.length / 2)
                {
                    return 0;
                }
                
                scancodes[0] = scancodes2[2 * indexScancodes];
                scancodes[1] = scancodes2[2 * indexScancodes + 1];
            }
            else
            {
                return 0;
            }
            
            /* Successful return, all failures return a 0 above. */
            return numScancodes;
        }
        
        public function Scancode (scancodes:ByteArray, event:KeyboardEvent, pressed:Boolean):uint
        {
            /* Convert the  event to a scancode sequence of a US keyboard,
             * because the client claims it is using the US one. Note: it does not
             * matter for the server because the client will send scancodes, and the
             * server will translate scancodes to the server keyboard layout.
             *
             * Return how many bytes were written to the scancodes array.
             *
             * @todo Take into account the current client keyboard layout: Y/Z US/DE?
             */
            var numScancodes:uint = scancode (scancodes, event, mKeyCodeTable);

            /* Workaround for Left/Right key mismatch.
             * Flash for example report the Right-Shift down/up as Right-Shift down + Left-Shift up.
             */
            if (numScancodes == 1)
            {
                if (scancodes[0] == 0x36) /* RShift */
                {
                    mfRightShiftPressed = pressed;
                }
                else if (scancodes[0] == 0x2a) /* LShift */
                {
                    if (!pressed && mfRightShiftPressed)
                    {
                        scancodes[0] = 0x36;
                        mfRightShiftPressed = false;
                    }
                }
            }

            return numScancodes;
        }
        
        public function SupportedLayouts():String
        {
            return "en,de";
        }
        
        public function CurrentLayout():String
        {
            return mCurrentLayout;
        }
        
        public function SetLayout(layout:String):void
        {
            layout = layout.toLowerCase();

            if (layout == "de")
            {
                mKeyCodeTable = keyCodeDE;
                mCurrentLayout = layout;
            }
            else
            {
                /* Everything else is EN. */
                mKeyCodeTable = keyCodeUS;
                mCurrentLayout = layout;
            }
            mClient.debug().msg("KeyboardTranslator: SetLayout req [" + layout + "], set [" + mCurrentLayout + "]\n");
            return;
        }
    }
}
